<?php
/* Copyright (C) 2004-2005  Rodolphe Quiedeville    <rodolphe@quiedeville.org>
 * Copyright (C) 2013       Olivier Geffroy         <jeff@jeffinfo.com>
 * Copyright (C) 2013-2019  Alexandre Spangaro      <aspangaro@open-dsi.fr>
 * Copyright (C) 2018       Frédéric France         <frederic.france@netlogic.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <https://www.gnu.org/licenses/>.
 */

/**
 * \file      	htdocs/accountancy/class/lettering.class.php
 * \ingroup 	Accountancy (Double entries)
 * \brief 		File of class for lettering
 */

include_once DOL_DOCUMENT_ROOT."/accountancy/class/bookkeeping.class.php";
include_once DOL_DOCUMENT_ROOT."/societe/class/societe.class.php";
include_once DOL_DOCUMENT_ROOT."/core/lib/date.lib.php";

/**
 * Class Lettering
 */
class Lettering extends BookKeeping
{
	/**
	 * letteringThirdparty
	 *
	 * @param int $socid Thirdparty id
	 * @return int 1 OK, <0 error
	 */
	public function letteringThirdparty($socid)
	{
		global $conf;

		$error = 0;

		$object = new Societe($this->db);
		$object->id = $socid;
		$object->fetch($socid);


		if ($object->code_compta == '411CUSTCODE') {
			$object->code_compta = '';
		}

		if ($object->code_compta_fournisseur == '401SUPPCODE') {
			$object->code_compta_fournisseur = '';
		}

		/**
		 * Prise en charge des lettering complexe avec prelevment , virement
		 */
		$sql = "SELECT DISTINCT bk.rowid, bk.doc_date, bk.doc_type, bk.doc_ref, bk.subledger_account, ";
		$sql .= " bk.numero_compte , bk.label_compte, bk.debit , bk.credit, bk.montant ";
		$sql .= " , bk.sens , bk.code_journal , bk.piece_num, bk.date_lettering, bu.url_id , bu.type ";
		$sql .= " FROM ".MAIN_DB_PREFIX."accounting_bookkeeping as bk";
		$sql .= " LEFT JOIN  ".MAIN_DB_PREFIX."bank_url as bu ON(bk.fk_doc = bu.fk_bank AND bu.type IN ('payment', 'payment_supplier') ) ";
		$sql .= " WHERE ( ";
		if ($object->code_compta != "") {
			$sql .= " bk.subledger_account = '".$this->db->escape($object->code_compta)."'  ";
		}
		if ($object->code_compta != "" && $object->code_compta_fournisseur != "") {
			$sql .= " OR ";
		}
		if ($object->code_compta_fournisseur != "") {
			$sql .= " bk.subledger_account = '".$this->db->escape($object->code_compta_fournisseur)."' ";
		}

		$sql .= " ) AND (bk.date_lettering ='' OR bk.date_lettering IS NULL) ";
		$sql .= "  AND (bk.lettering_code != '' OR bk.lettering_code IS NULL) ";
		$sql .= ' AND bk.date_validated IS NULL ';
		$sql .= $this->db->order('bk.doc_date', 'DESC');

		// echo $sql;
		//
		$resql = $this->db->query($sql);
		if ($resql) {
			$num = $this->db->num_rows($resql);

			while ($obj = $this->db->fetch_object($resql)) {
				$ids = array();
				$ids_fact = array();

				if ($obj->type == 'payment_supplier') {
					$sql = 'SELECT DISTINCT bk.rowid, facf.ref, facf.ref_supplier, payf.fk_bank, facf.rowid as fact_id';
					$sql .= " FROM ".MAIN_DB_PREFIX."facture_fourn facf ";
					$sql .= " INNER JOIN ".MAIN_DB_PREFIX."paiementfourn_facturefourn as payfacf ON  payfacf.fk_facturefourn=facf.rowid";
					$sql .= " INNER JOIN ".MAIN_DB_PREFIX."paiementfourn as payf ON  payfacf.fk_paiementfourn=payf.rowid";
					$sql .= " INNER JOIN ".MAIN_DB_PREFIX."accounting_bookkeeping as bk ON (bk.fk_doc = payf.fk_bank AND bk.code_journal='".$this->db->escape($obj->code_journal)."')";
					$sql .= " WHERE payfacf.fk_paiementfourn = '".$this->db->escape($obj->url_id)."' ";
					$sql .= " AND facf.entity = ".$conf->entity;
					$sql .= " AND code_journal IN (SELECT code FROM ".MAIN_DB_PREFIX."accounting_journal WHERE nature=4 AND entity=".$conf->entity.") ";
					$sql .= " AND ( ";
					if ($object->code_compta != "") {
						$sql .= "  bk.subledger_account = '".$this->db->escape($object->code_compta)."'  ";
					}
					if ($object->code_compta != "" && $object->code_compta_fournisseur != "") {
						$sql .= "  OR  ";
					}
					if ($object->code_compta_fournisseur != "") {
						$sql .= "   bk.subledger_account = '".$this->db->escape($object->code_compta_fournisseur)."' ";
					}
					$sql .= " )  ";

					$resql2 = $this->db->query($sql);
					if ($resql2) {
						while ($obj2 = $this->db->fetch_object($resql2)) {
							$ids[$obj2->rowid] = $obj2->rowid;
							$ids_fact[] = $obj2->fact_id;
						}
						$this->db->free($resql2);
					} else {
						$this->errors[] = $this->db->lasterror;
						return -1;
					}
					if (count($ids_fact)) {
						$sql = 'SELECT bk.rowid, facf.ref, facf.ref_supplier ';
						$sql .= " FROM ".MAIN_DB_PREFIX."facture_fourn facf ";
						$sql .= " INNER JOIN ".MAIN_DB_PREFIX."accounting_bookkeeping as bk ON(  bk.fk_doc = facf.rowid AND facf.rowid IN (".$this->db->sanitize(implode(',', $ids_fact))."))";
						$sql .= " WHERE bk.code_journal IN (SELECT code FROM ".MAIN_DB_PREFIX."accounting_journal WHERE nature=3 AND entity=".$conf->entity.") ";
						$sql .= " AND facf.entity = ".$conf->entity;
						$sql .= " AND ( ";
						if ($object->code_compta != "") {
							$sql .= " bk.subledger_account = '".$this->db->escape($object->code_compta)."'  ";
						}
						if ($object->code_compta != "" && $object->code_compta_fournisseur != "") {
							$sql .= " OR ";
						}
						if ($object->code_compta_fournisseur != "") {
							$sql .= " bk.subledger_account = '".$this->db->escape($object->code_compta_fournisseur)."' ";
						}
						$sql .= ") ";

						$resql2 = $this->db->query($sql);
						if ($resql2) {
							while ($obj2 = $this->db->fetch_object($resql2)) {
								$ids[$obj2->rowid] = $obj2->rowid;
							}
							$this->db->free($resql2);
						} else {
							$this->errors[] = $this->db->lasterror;
							return -1;
						}
					}
				} elseif ($obj->type == 'payment') {
					$sql = 'SELECT DISTINCT bk.rowid, fac.ref, fac.ref, pay.fk_bank, fac.rowid as fact_id';
					$sql .= " FROM ".MAIN_DB_PREFIX."facture fac ";
					$sql .= " INNER JOIN ".MAIN_DB_PREFIX."paiement_facture as payfac ON  payfac.fk_facture=fac.rowid";
					$sql .= " INNER JOIN ".MAIN_DB_PREFIX."paiement as pay ON  payfac.fk_paiement=pay.rowid";
					$sql .= " INNER JOIN ".MAIN_DB_PREFIX."accounting_bookkeeping as bk ON (bk.fk_doc = pay.fk_bank AND bk.code_journal='".$this->db->escape($obj->code_journal)."')";
					$sql .= " WHERE payfac.fk_paiement = '".$this->db->escape($obj->url_id)."' ";
					$sql .= " AND bk.code_journal IN (SELECT code FROM ".MAIN_DB_PREFIX."accounting_journal WHERE nature=4 AND entity=".$conf->entity.") ";
					$sql .= " AND fac.entity IN (".getEntity('invoice', 0).")"; // We don't share object for accountancy
					$sql .= " AND ( ";
					if ($object->code_compta != "") {
						$sql .= "  bk.subledger_account = '".$this->db->escape($object->code_compta)."'  ";
					}
					if ($object->code_compta != "" && $object->code_compta_fournisseur != "") {
						$sql .= "  OR  ";
					}
					if ($object->code_compta_fournisseur != "") {
						$sql .= "   bk.subledger_account = '".$this->db->escape($object->code_compta_fournisseur)."' ";
					}
					$sql .= " )";

					$resql2 = $this->db->query($sql);
					if ($resql2) {
						while ($obj2 = $this->db->fetch_object($resql2)) {
							$ids[$obj2->rowid] = $obj2->rowid;
							$ids_fact[] = $obj2->fact_id;
						}
					} else {
						$this->errors[] = $this->db->lasterror;
						return -1;
					}
					if (count($ids_fact)) {
						$sql = 'SELECT bk.rowid, fac.ref, fac.ref_supplier ';
						$sql .= " FROM ".MAIN_DB_PREFIX."facture fac ";
						$sql .= " INNER JOIN ".MAIN_DB_PREFIX."accounting_bookkeeping as bk ON(  bk.fk_doc = fac.rowid AND fac.rowid IN (".$this->db->sanitize(implode(',', $ids_fact))."))";
						$sql .= " WHERE code_journal IN (SELECT code FROM ".MAIN_DB_PREFIX."accounting_journal WHERE nature=2 AND entity=".$conf->entity.") ";
						$sql .= " AND fac.entity IN (".getEntity('invoice', 0).")"; // We don't share object for accountancy
						$sql .= " AND ( ";
						if ($object->code_compta != "") {
							$sql .= "  bk.subledger_account = '".$this->db->escape($object->code_compta)."'  ";
						}
						if ($object->code_compta != "" && $object->code_compta_fournisseur != "") {
							$sql .= "  OR  ";
						}
						if ($object->code_compta_fournisseur != "") {
							$sql .= "   bk.subledger_account = '".$this->db->escape($object->code_compta_fournisseur)."' ";
						}
						$sql .= " )  ";

						$resql2 = $this->db->query($sql);
						if ($resql2) {
							while ($obj2 = $this->db->fetch_object($resql2)) {
								$ids[$obj2->rowid] = $obj2->rowid;
							}
							$this->db->free($resql2);
						} else {
							$this->errors[] = $this->db->lasterror;
							return -1;
						}
					}
				}

				if (count($ids) > 1) {
					$result = $this->updateLettering($ids);
				}
			}
			$this->db->free($resql);
		}
		if ($error) {
			foreach ($this->errors as $errmsg) {
				dol_syslog(__METHOD__.' '.$errmsg, LOG_ERR);
				$this->error .= ($this->error ? ', '.$errmsg : $errmsg);
			}
			return -1 * $error;
		} else {
			return 1;
		}
	}

	/**
	 *
	 * @param	array		$ids			ids array
	 * @param	int			$company_id		Company ID
	 * @param	boolean		$notrigger		no trigger
	 * @return	int
	 */
	public function updateLettering($ids, $company_id, $notrigger = false)
	{
		$error = 0;
		$lettre = 'AAA';

		$sql = "SELECT DISTINCT ab.lettering_code" .
			" FROM " . MAIN_DB_PREFIX . "accounting_bookkeeping As ab" .
			" LEFT JOIN " . MAIN_DB_PREFIX . "bank_url As bu ON bu.fk_bank = ab.fk_doc" .
			" WHERE ab.doc_type = 'bank'" .
			" AND bu.type = 'company'" .
			" AND ab.subledger_account != ''" .
			" AND ab.lettering_code != ''" .
			" AND bu.url_id = " . ((int) $company_id) .
			" ORDER BY ab.lettering_code DESC" .
			" LIMIT 1 ";

		$result = $this->db->query($sql);
		if ($result) {
			$obj = $this->db->fetch_object($result);
			$lettre = (empty($obj->lettering_code) ? 'AAA' : $obj->lettering_code);
			if (!empty($obj->lettering_code)) {
				$lettre++;
			}
			$this->db->free($result);
		} else {
			$this->errors[] = 'Error'.$this->db->lasterror();
			$error++;
		}

		$sql = "SELECT SUM(ABS(debit)) as deb, SUM(ABS(credit)) as cred FROM ".MAIN_DB_PREFIX."accounting_bookkeeping WHERE ";
		$sql .= " rowid IN (".$this->db->sanitize(implode(',', $ids)).") AND lettering_code IS NULL AND subledger_account != ''";
		$result = $this->db->query($sql);
		if ($result) {
			$obj = $this->db->fetch_object($result);
			if (!(round(abs($obj->deb), 2) === round(abs($obj->cred), 2))) {
				$this->errors[] = 'Total not exacts '.round(abs($obj->deb), 2).' vs '.round(abs($obj->cred), 2);
				$error++;
			}
			$this->db->free($result);
		} else {
			$this->errors[] = 'Erreur sql'.$this->db->lasterror();
			$error++;
		}

		// Update request

		$now = dol_now();

		if (!$error) {
			$sql = "UPDATE ".MAIN_DB_PREFIX."accounting_bookkeeping SET";
			$sql .= " lettering_code='".$this->db->escape($lettre)."'";
			$sql .= " , date_lettering = '".$this->db->idate($now)."'"; // todo correct date it's false
			$sql .= "  WHERE rowid IN (".$this->db->sanitize(implode(',', $ids)).") AND lettering_code IS NULL AND subledger_account != ''";

			dol_syslog(get_class($this)."::update", LOG_DEBUG);
			$resql = $this->db->query($sql);
			if (!$resql) {
				$error++;
				$this->errors[] = "Error ".$this->db->lasterror();
			}
		}

		// Commit or rollback
		if ($error) {
			foreach ($this->errors as $errmsg) {
				dol_syslog(get_class($this)."::update ".$errmsg, LOG_ERR);
				$this->error .= ($this->error ? ', '.$errmsg : $errmsg);
			}
			return -1 * $error;
		} else {
			return 1;
		}
	}

	/**
	 * Lettering bookkeeping lines all types
	 *
	 * @param	string		$subledger_account		Lettering specific subledger account
	 * @param	int			$company_id				Lettering specific company
	 * @param	array		$bookkeeping_ids		Lettering specific list of bookkeeping id
	 * @return	int									<0 if error (nb lettered = result -1), 0 if noting to lettering, >0 if OK (nb lettered)
	 */
	public function bookkeepingLetteringAll($subledger_account = '', $company_id = 0, $bookkeeping_ids = array())
	{
		dol_syslog(__METHOD__ . " - ", LOG_DEBUG);

		$error = 0;
		$errors = array();
		$nb_lettering = 0;

		$result = $this->bookkeepingLettering('customer_invoice', $subledger_account, $company_id, $bookkeeping_ids);
		if ($result < 0) {
			$error++;
			$errors = array_merge($errors, $this->errors);
			$nb_lettering += abs($result) - 2;
		} else {
			$nb_lettering += $result;
		}

		$result = $this->bookkeepingLettering('supplier_invoice', $subledger_account, $company_id, $bookkeeping_ids);
		if ($result < 0) {
			$error++;
			$errors = array_merge($errors, $this->errors);
			$nb_lettering += abs($result) - 2;
		} else {
			$nb_lettering += $result;
		}

		if ($error) {
			$this->errors = $errors;
			return -2 - $nb_lettering;
		} else {
			return $nb_lettering;
		}
	}

	/**
	 * Lettering bookkeeping lines
	 *
	 * @param	string		$type					Type of bookkeeping type to lettering ('customer_invoice' or 'supplier_invoice')
	 * @param	string		$subledger_account		Lettering specific subledger account
	 * @param	int			$company_id				Lettering specific company
	 * @param	array		$bookkeeping_ids		Lettering specific list of bookkeeping id
	 * @return	int									<0 if error (nb lettered = result -1), 0 if noting to lettering, >0 if OK (nb lettered)
	 */
	public function bookkeepingLettering($type = 'customer_invoice', $subledger_account = '', $company_id = 0, $bookkeeping_ids = array())
	{
		global $conf, $langs;
		$this->errors = array();

		// Clean parameters
		$type = trim($type);
		$subledger_account = trim($subledger_account);
		$company_id = $company_id > 0 ? $company_id : 0;
		$bookkeeping_ids = is_array($bookkeeping_ids) ? $bookkeeping_ids : array();

		if ($type == 'customer_invoice') {
			$doc_type = 'customer_invoice';
			$bank_url_type = 'payment';
			$payment_element = 'paiement_facture';
			$fk_payment_element = 'fk_paiement';
			$element = 'facture';
			$fk_element = 'fk_facture';
			$account_number = $conf->global->ACCOUNTING_ACCOUNT_CUSTOMER;
		} elseif ($type == 'supplier_invoice') {
			$doc_type = 'supplier_invoice';
			$bank_url_type = 'payment_supplier';
			$payment_element = 'paiementfourn_facturefourn';
			$fk_payment_element = 'fk_paiementfourn';
			$element = 'facture_fourn';
			$fk_element = 'fk_facturefourn';
			$account_number = $conf->global->ACCOUNTING_ACCOUNT_SUPPLIER;
		} else {
			$langs->load('errors');
			$this->errors[] = $langs->trans('ErrorBadParameters');
			return -1;
		}

		$bank_by_element = array();
		$element_by_bank = array();
		$lines = array();

		// Get bank lines with element infos
		$sql = "SELECT DISTINCT ab.fk_doc AS bank_id, e.rowid AS element_id, e.fk_soc AS company_id" .
			", ab.rowid AS bank_bookkeeping_id, ab.piece_num AS bank_piece_num, ab.subledger_account AS bank_subledger_account, IF(ab.sens = 'C', 1, -1) * pe.amount AS bank_amount" .
			", ab2.rowid AS element_bookkeeping_id, ab2.piece_num AS element_piece_num, ab2.subledger_account AS element_subledger_account, IF(ab2.credit > 0, ab2.credit, -1 * ab2.debit) AS element_amount" .
			" FROM " . MAIN_DB_PREFIX . "accounting_bookkeeping As ab" .
			" LEFT JOIN " . MAIN_DB_PREFIX . "bank_url As bu ON bu.fk_bank = ab.fk_doc" .
			" LEFT JOIN " . MAIN_DB_PREFIX . "$payment_element AS pe ON pe.$fk_payment_element = bu.url_id" .
			" LEFT JOIN " . MAIN_DB_PREFIX . "$element AS e ON e.rowid = pe.$fk_element" .
			" LEFT JOIN " . MAIN_DB_PREFIX . "accounting_bookkeeping AS ab2 ON ab2.fk_doc = e.rowid" .
			" WHERE ab.doc_type = 'bank'" .
			" AND ab2.doc_type = '" . $this->db->escape($doc_type) . "'" .
			" AND ab.numero_compte = '" . $this->db->escape($account_number) . "'" .
			" AND ab2.numero_compte = '" . $this->db->escape($account_number) . "'" .
			" AND ab.lettering_code IS NULL" .
			" AND ab2.lettering_code IS NULL" .
			" AND bu.type = '" . $this->db->escape($bank_url_type) . "'";
		if (empty($subledger_account)) {
			$sql .= " AND ab.subledger_account != ''";
			$sql .= " AND ab2.subledger_account != ''";
		} else {
			$sql .= " AND ab.subledger_account = '" . $this->db->escape($subledger_account) . "'";
			$sql .= " AND ab2.subledger_account = '" . $this->db->escape($subledger_account) . "'";
		}
		if ($company_id > 0) $sql .= " AND e.fk_soc = " . $company_id;
		if (!empty($bookkeeping_ids)) $sql .= " AND ab.rowid IN (" . $this->db->sanitize(implode(',', $bookkeeping_ids)) . ")";

		dol_syslog(__METHOD__ . " - Get bank lines", LOG_DEBUG);
		$resql = $this->db->query($sql);
		if (!$resql) {
			$this->errors[] = "Error " . $this->db->lasterror();
			return -1;
		}

		if ($this->db->num_rows($resql) == 0) {
			// Nothing to lettering
			return 0;
		}

		while ($obj = $this->db->fetch_object($resql)) {
			$bank_by_element[$obj->element_id][$obj->bank_id] = $obj->bank_id;
			$element_by_bank[$obj->bank_id][$obj->element_id] = $obj->element_id;
			$lines[$obj->bank_id][$obj->element_id] = array(
				'bank_id' => $obj->bank_id,
				'element_id' => $obj->element_id,
				'company_id' => $obj->company_id,
				'bank_bookkeeping_id' => $obj->bank_bookkeeping_id,
				'bank_piece_num' => $obj->bank_piece_num,
				'bank_subledger_account' => $obj->bank_subledger_account,
				'bank_amount' => $obj->bank_amount,
				'element_bookkeeping_id' => $obj->element_bookkeeping_id,
				'element_piece_num' => $obj->element_piece_num,
				'element_subledger_account' => $obj->element_subledger_account,
				'element_amount' => $obj->element_amount,
			);
		}
		$this->db->free($resql);

		$error = 0;
		$nb_lettering = 0;
		$grouped_elements = $this->getGroupElements($bank_by_element, $element_by_bank);
		foreach ($grouped_elements as $element_ids) {
			$group_error = 0;
			$total = 0;
			$company_id = 0;
			$subledger_account = '';
			$piece_num_lines = array();
			$bookkeeping_lines = array();
			foreach ($element_ids as $element_id) {
				foreach ($bank_by_element[$element_id] as $bank_id) {
					$line = $lines[$bank_id][$element_id];

					$piece_num_lines[$line['bank_piece_num']] = $line['bank_piece_num'];
					$piece_num_lines[$line['element_piece_num']] = $line['element_piece_num'];
					$bookkeeping_lines[$line['bank_bookkeeping_id']] = $line['bank_bookkeeping_id'];
					$bookkeeping_lines[$line['element_bookkeeping_id']] = $line['element_bookkeeping_id'];
					$total += $line['bank_amount'];

					// Check subledger account
					if ((!empty($subledger_account) && $subledger_account != $line['bank_subledger_account']) ||
						$line['bank_subledger_account'] != $line['element_subledger_account']
					) {
						$this->errors[] = $langs->trans('AccountancyErrorMismatchSubLedgerAccount');
						$group_error++;
						break;
					} elseif (empty($subledger_account)) {
						$subledger_account = $line['bank_subledger_account'];
					}

					// Check company id
					if ($company_id > 0 && $company_id != $line['company_id']) {
						$this->errors[] = $langs->trans('AccountancyErrorMismatchCompanyId');
						$group_error++;
						break;
					} elseif (empty($company_id)) {
						$company_id = $line['company_id'];
					}
				}
				if (isset($line)) $total += $line['element_amount'];
				if ($group_error) break;
			}

			// Check balance amount
			if (!$group_error && price2num($total) != 0) {
				$this->errors[] = $langs->trans('AccountancyErrorMismatchBalanceAmount', $total);
				$group_error++;
			}

			// Lettering the group of bookkeeping lines
			if (!$group_error) {
				$result = $this->updateLettering($bookkeeping_lines, $company_id);
				if ($result < 0) {
					$group_error++;
				}
			}

			if ($group_error) {
				$this->errors[] = $langs->trans('AccountancyErrorLetteringBookkeeping', implode(', ', $piece_num_lines));
				$error++;
			} else {
				$nb_lettering++;
			}
		}

		if ($error) {
			return -2 - $nb_lettering;
		} else {
			return $nb_lettering;
		}
	}

	/**
	 * Get elements ids grouped by bank id and element id in common
	 *
	 * @param	array	&$bank_by_element	List of bank ids by element id
	 * @param	array	&$element_by_bank	List of element ids by bank id
	 * @param	int		$bank_id			Bank Id (used for recursive function)
	 * @param	array	&$current_group		Current group (used for recursive function)
	 * @return	array						List of elements ids grouped by bank id and element id in common
	 */
	public function getGroupElements(&$bank_by_element, &$element_by_bank, $bank_id = 0, &$current_group = array())
	{
		$grouped_elements = array();
		if ($bank_id > 0 && !isset($element_by_bank[$bank_id])) {
			// Return if specific bank id not found
			return $grouped_elements;
		}

		if ($bank_id == 0) {
			// Save list when is the begin of recursive function
			$save_bank_by_element = $bank_by_element;
			$save_element_by_bank = $element_by_bank;
		}

		do {
			// Get current bank id, get this element id list and delete the entry
			$current_bank_id = $bank_id > 0 ? $bank_id : array_keys($element_by_bank)[0];
			$element_ids = $element_by_bank[$current_bank_id];
			unset($element_by_bank[$current_bank_id]);

			foreach ($element_ids as $element_id) {
				// Continue if element id in not found
				if (!isset($bank_by_element[$element_id])) continue;

				// Set the element in the current group
				$current_group[$element_id] = $element_id;

				// Get current bank id, get this element id list and delete the entry
				$bank_ids = $bank_by_element[$element_id];
				unset($bank_by_element[$element_id]);

				// Set element id on the current group for each bank id of the element
				foreach ($bank_ids as $id) {
					$this->getGroupElements($bank_by_element, $element_by_bank, $id, $current_group);
				}
			}

			if ($bank_id == 0) {
				// Save current group and reset the current group when is the begin of recursive function
				$grouped_elements[] = $current_group;
				$current_group = array();
			}
		} while(!empty($element_by_bank) && $bank_id == 0);

		if ($bank_id == 0) {
			// Restore list when is the begin of recursive function
			$bank_by_element = $save_bank_by_element;
			$element_by_bank = $save_element_by_bank;
		}

		return $grouped_elements;
	}
}
